// This contains various minified Javascript libraries.
// Since I don't plan to use a lot of them, it's simpler to do things manually
// than to automate it.
// Licensing varies.
#pragma once
#include <rack.hpp>


namespace JavascriptLibraries {

// Snippets from https://github.com/AriaSalvatrice/js-snippets-for-modules
// WTFPL
static const std::string& TOKENIZE = R"END(
function tokenize(input) {
    input = input.replace("(", "") // No parentheses
    input = input.replace(")", "")
    input = input.split(/[,-\s]\s*/) // Commas, hyphens, spaces
    input = input.filter(element => {return element != ""})
    return input
}
)END";

static const std::string& TOVOCT = R"END(
// MIDI 60 = 0V = C4 = 261.6256 hz
function toVOct(pitch){
    pitch = Tonal.Midi.toMidi(pitch)
    pitch = (pitch - 60) * 1 / 12
    return pitch
}
)END";

static const std::string& TOSCALEPOSITION = R"END(
function toScalePosition(pitch){
    pitch = Tonal.Midi.toMidi(pitch)
    pitch = (pitch - 60)
    if(pitch >= 12) pitch = pitch - 12
    return pitch
}
)END";

static const std::string& PARSEASLEADSHEET = R"END(
// Tonic on the 4th octave for easier further conversion
function parseAsLeadsheet(input) {
    pitchSeries = input.map(currentChord => {       
        if (currentChord.includes("/")) {
            split = currentChord.split("/")
            currentChord = Tonal.Chord.get(split[0])
            currentChord = Tonal.Chord.getChord(currentChord.aliases[0], currentChord.tonic, split[1])
        } else {
            currentChord = Tonal.Chord.get(currentChord)
        }
        currentFirstNote = Tonal.note(currentChord.tonic + "4")
        return currentChord.intervals.map(interval => Tonal.transpose(currentFirstNote, interval))

    })
    return pitchSeries
}
)END";

static const std::string& LEADSHEETTOQQQQ = R"END(
function leadsheetToQqqq(input) {
    results = parseAsLeadsheet(tokenize(input))
    results = results.map(chord => chord.map(pitch => toScalePosition(pitch)))
    return JSON.stringify(results)
}
)END";

static const std::string& ROMANTOQQQQ = R"END(
function romanToQqqq(tonic, input) {
    input = tokenize(input)
    progression = Tonal.Progression.fromRomanNumerals(tonic, input)
    results = parseAsLeadsheet(progression)
    results = results.map(chord => chord.map(pitch => toScalePosition(pitch)))
    return JSON.stringify(results)
}
)END";




///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////




// Tonal.js version 4.0.1
// https://github.com/tonaljs/tonal
// Using tonal.min.js as provided by the project
//
//
// The MIT License (MIT)
// 
// Copyright (c) 2015 danigb
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
static const std::string& TONALJS = R"END(
var Tonal=function(){"use strict";"undefined"!=typeof globalThis?globalThis:"undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self&&self;const t=(t,n)=>Array(Math.abs(n)+1).join(t);function n(t,n,e){return function(...o){return console.warn(`${t} is deprecated. Use ${n}.`),e.apply(this,o)}}function e(t){return null!==t&&"object"==typeof t&&"string"==typeof t.name}function o(t){return null!==t&&"object"==typeof t&&"number"==typeof t.step&&"number"==typeof t.alt}const r=[0,2,4,-1,1,3,5],a=r.map(t=>Math.floor(7*t/12));function m(t){const{step:n,alt:e,oct:o,dir:m=1}=t,i=r[n]+7*e;return void 0===o?[m*i]:[m*i,m*(o-a[n]-4*e)]}const i=[3,0,4,1,5,2,6];function s(t){const[n,e,o]=t,r=i[function(t){const n=(t+1)%7;return n<0?7+n:n}(n)],m=Math.floor((n+1)/7);return void 0===e?{step:r,alt:m,dir:o}:{step:r,alt:m,oct:e+4*m+a[r],dir:o}}const c={empty:!0,name:"",pc:"",acc:""},u=new Map,l=t=>"CDEFGAB".charAt(t),P=n=>n<0?t("b",-n):t("#",n),d=t=>"b"===t[0]?-t.length:t.length;function M(t){const n=u.get(t);if(n)return n;const r="string"==typeof t?function(t){const n=f(t);if(""===n[0]||""!==n[3])return c;const e=n[0],o=n[1],r=n[2],a=(e.charCodeAt(0)+3)%7,i=d(o),s=r.length?+r:void 0,u=m({step:a,alt:i,oct:s}),l=e+o+r,P=e+o,M=(y[a]+i+120)%12,p=void 0===s?-100:s,h=y[a]+i+12*(p+1),b=h>=0&&h<=127?h:null,g=void 0===s?null:440*Math.pow(2,(h-69)/12);return{empty:!1,acc:o,alt:i,chroma:M,coord:u,freq:g,height:h,letter:e,midi:b,name:l,oct:s,pc:P,step:a}}(t):o(t)?M(function(t){const{step:n,alt:e,oct:o}=t,r=l(n);if(!r)return"";const a=r+P(e);return o||0===o?a+o:a}(t)):e(t)?M(t.name):c;return u.set(t,r),r}const p=/^([a-gA-G]?)(#{1,}|b{1,}|x{1,}|)(-?\d*)\s*(.*)$/;function f(t){const n=p.exec(t);return[n[1].toUpperCase(),n[2].replace(/x/g,"##"),n[3],n[4]]}function h(t){return M(s(t))}const y=[0,2,4,5,7,9,11];const b={empty:!0,name:"",acc:""},g=new RegExp("^([-+]?\\d+)(d{1,4}|m|M|P|A{1,4})|(AA|A|P|M|m|d|dd)([-+]?\\d+)$");function A(t){const n=g.exec(""+t);return null===n?["",""]:n[1]?[n[1],n[2]]:[n[4],n[3]]}const v={};function j(n){return"string"==typeof n?v[n]||(v[n]=function(t){const n=A(t);if(""===n[0])return b;const e=+n[0],o=n[1],r=(Math.abs(e)-1)%7,a="PMMPPMM"[r];if("M"===a&&"P"===o)return b;const i="M"===a?"majorable":"perfectable",s=""+e+o,c=e<0?-1:1,u=8===e||-8===e?e:c*(r+1),l=function(t,n){return"M"===n&&"majorable"===t||"P"===n&&"perfectable"===t?0:"m"===n&&"majorable"===t?-1:/^A+$/.test(n)?n.length:/^d+$/.test(n)?-1*("perfectable"===t?n.length:n.length+1):0}(i,o),P=Math.floor((Math.abs(e)-1)/7),d=c*(I[r]+l+12*P),M=(c*(I[r]+l)%12+12)%12,p=m({step:r,alt:l,oct:P,dir:c});return{empty:!1,name:s,num:e,q:o,step:r,alt:l,dir:c,type:i,simple:u,semitones:d,chroma:M,coord:p,oct:P}}(n)):o(n)?j(function(n){const{step:e,alt:o,oct:r=0,dir:a}=n;if(!a)return"";return(a<0?"-":"")+(e+1+7*r)+function(n,e){return 0===e?"majorable"===n?"M":"P":-1===e&&"majorable"===n?"m":e>0?t("A",e):t("d","perfectable"===n?e:e+1)}("M"==="PMMPPMM"[e]?"majorable":"perfectable",o)}(n)):e(n)?j(n.name):b}const I=[0,2,4,5,7,9,11];function w(t){const[n,e=0]=t;return j(s(7*n+12*e<0?[-n,-e,-1]:[n,e,1]))}function O(t,n){const e=M(t),o=j(n);if(e.empty||o.empty)return"";const r=e.coord,a=o.coord;return h(1===r.length?[r[0]+a[0]]:[r[0]+a[0],r[1]+a[1]]).name}function N(t,n){const e=M(t),o=M(n);if(e.empty||o.empty)return"";const r=e.coord,a=o.coord,m=a[0]-r[0];return w([m,2===r.length&&2===a.length?a[1]-r[1]:-Math.floor(7*m/12)]).name}var x=Object.freeze({__proto__:null,accToAlt:d,altToAcc:P,coordToInterval:w,coordToNote:h,decode:s,deprecate:n,distance:N,encode:m,fillStr:t,interval:j,isNamed:e,isPitch:o,note:M,stepToLetter:l,tokenizeInterval:A,tokenizeNote:f,transpose:O});const T=(t,n)=>Array(n+1).join(t),D=/^(_{1,}|=|\^{1,}|)([abcdefgABCDEFG])([,']*)$/;function S(t){const n=D.exec(t);return n?[n[1],n[2],n[3]]:["","",""]}function V(t){const[n,e,o]=S(t);if(""===e)return"";let r=4;for(let t=0;t<o.length;t++)r+=","===o.charAt(t)?-1:1;const a="_"===n[0]?n.replace(/_/g,"b"):"^"===n[0]?n.replace(/\^/g,"#"):"";return e.charCodeAt(0)>96?e.toUpperCase()+a+(r+1):e+a+r}function C(t){const n=M(t);if(n.empty||!n.oct)return"";const{letter:e,acc:o,oct:r}=n;return("b"===o[0]?o.replace(/b/g,"_"):o.replace(/#/g,"^"))+(r>4?e.toLowerCase():e)+(5===r?"":r>4?T("'",r-5):T(",",4-r))}var $={abcToScientificNotation:V,scientificToAbcNotation:C,tokenize:S,transpose:function(t,n){return C(O(V(t),n))},distance:function(t,n){return N(V(t),V(n))}};function q(t){return t.map(t=>M(t)).filter(t=>!t.empty).sort((t,n)=>t.height-n.height).map(t=>t.name)}var k=Object.freeze({__proto__:null,compact:function(t){return t.filter(t=>0===t||t)},permutations:function t(n){return 0===n.length?[[]]:t(n.slice(1)).reduce((t,e)=>t.concat(n.map((t,o)=>{const r=e.slice();return r.splice(o,0,n[0]),r})),[])},range:function(t,n){return t<n?function(t,n){const e=[];for(;n--;e[n]=n+t);return e}(t,n-t+1):function(t,n){const e=[];for(;n--;e[n]=t-n);return e}(t,t-n+1)},rotate:function(t,n){const e=n.length,o=(t%e+e)%e;return n.slice(o,e).concat(n.slice(0,o))},shuffle:function(t,n=Math.random){let e,o,r=t.length;for(;r;)e=Math.floor(n()*r--),o=t[r],t[r]=t[e],t[e]=o;return t},sortedNoteNames:q,sortedUniqNoteNames:function(t){return q(t).filter((t,n,e)=>0===n||t!==e[n-1])}});function E(t,n){return t<n?function(t,n){const e=[];for(;n--;e[n]=n+t);return e}(t,n-t+1):function(t,n){const e=[];for(;n--;e[n]=t-n);return e}(t,t-n+1)}function _(t,n){const e=n.length,o=(t%e+e)%e;return n.slice(o,e).concat(n.slice(0,o))}function F(t){return t.filter(t=>0===t||t)}var z={compact:F,permutations:function t(n){return 0===n.length?[[]]:t(n.slice(1)).reduce((t,e)=>t.concat(n.map((t,o)=>{const r=e.slice();return r.splice(o,0,n[0]),r})),[])},range:E,rotate:_,shuffle:function(t,n=Math.random){let e,o,r=t.length;for(;r;)e=Math.floor(n()*r--),o=t[r],t[r]=t[e],t[e]=o;return t}};const R={empty:!0,name:"",setNum:0,chroma:"000000000000",normalized:"000000000000",intervals:[]},U=t=>Number(t).toString(2),B=t=>parseInt(t,2),G=/^[01]{12}$/;function K(t){return G.test(t)}const L={[R.chroma]:R};function H(t){const n=K(t)?t:"number"==typeof(e=t)&&e>=0&&e<=4095?U(t):Array.isArray(t)?function(t){if(0===t.length)return R.chroma;let n;const e=[0,0,0,0,0,0,0,0,0,0,0,0];for(let o=0;o<t.length;o++)n=M(t[o]),n.empty&&(n=j(t[o])),n.empty||(e[n.chroma]=1);return e.join("")}(t):(t=>t&&K(t.chroma))(t)?t.chroma:R.chroma;var e;return L[n]=L[n]||function(t){const n=B(t),e=function(t){const n=t.split("");return n.map((t,e)=>_(e,n).join(""))}(t).map(B).filter(t=>t>=2048).sort()[0],o=U(e),r=W(t);return{empty:!1,name:"",setNum:n,chroma:t,normalized:o,intervals:r}}(n)}const J=n("Pcset.pcset","Pcset.get",H),Q=["1P","2m","2M","3m","3M","4P","5d","5P","6m","6M","7m","7M"];function W(t){const n=[];for(let e=0;e<12;e++)"1"===t.charAt(e)&&n.push(Q[e]);return n}function X(t,n=!0){const e=H(t).chroma.split("");return F(e.map((t,o)=>{const r=_(o,e);return n&&"0"===r[0]?null:r.join("")}))}function Y(t){const n=H(t).setNum;return t=>{const e=H(t).setNum;return n&&n!==e&&(e&n)===e}}function Z(t){const n=H(t).setNum;return t=>{const e=H(t).setNum;return n&&n!==e&&(e|n)===e}}function tt(t){const n=H(t);return t=>{const e=M(t);return n&&!e.empty&&"1"===n.chroma.charAt(e.chroma)}}var nt={get:H,chroma:t=>H(t).chroma,num:t=>H(t).setNum,intervals:t=>H(t).intervals,chromas:function(){return E(2048,4095).map(U)},isSupersetOf:Z,isSubsetOf:Y,isNoteIncludedIn:tt,isEqual:function(t,n){return H(t).setNum===H(n).setNum},filter:function(t){const n=tt(t);return t=>t.filter(n)},modes:X,pcset:J};const et={...R,name:"",quality:"Unknown",intervals:[],aliases:[]};let ot=[],rt={};function at(t){return rt[t]||et}const mt=n("ChordType.chordType","ChordType.get",at);function it(){return ot.slice()}const st=n("ChordType.entries","ChordType.all",it);function ct(t,n,e){const o=function(t){const n=n=>-1!==t.indexOf(n);return n("5A")?"Augmented":n("3M")?"Major":n("5d")?"Diminished":n("3m")?"Minor":"Unknown"}(t),r={...H(t),name:e||"",quality:o,intervals:t,aliases:n};ot.push(r),r.name&&(rt[r.name]=r),rt[r.setNum]=r,rt[r.chroma]=r,r.aliases.forEach(t=>function(t,n){rt[n]=t}(r,t))}[["1P 3M 5P","major","M "],["1P 3M 5P 7M","major seventh","maj7 Δ ma7 M7 Maj7"],["1P 3M 5P 7M 9M","major ninth","maj9 Δ9"],["1P 3M 5P 7M 9M 13M","major thirteenth","maj13 Maj13"],["1P 3M 5P 6M","sixth","6 add6 add13 M6"],["1P 3M 5P 6M 9M","sixth/ninth","6/9 69"],["1P 3M 5P 7M 11A","lydian","maj#4 Δ#4 Δ#11"],["1P 3M 6m 7M","major seventh flat sixth","M7b6"],["1P 3m 5P","minor","m min -"],["1P 3m 5P 7m","minor seventh","m7 min7 mi7 -7"],["1P 3m 5P 7M","minor/major seventh","m/ma7 m/maj7 mM7 mMaj7 m/M7 -Δ7 mΔ"],["1P 3m 5P 6M","minor sixth","m6"],["1P 3m 5P 7m 9M","minor ninth","m9"],["1P 3m 5P 7m 9M 11P","minor eleventh","m11"],["1P 3m 5P 7m 9M 13M","minor thirteenth","m13"],["1P 3m 5d","diminished","dim ° o"],["1P 3m 5d 7d","diminished seventh","dim7 °7 o7"],["1P 3m 5d 7m","half-diminished","m7b5 ø"],["1P 3M 5P 7m","dominant seventh","7 dom"],["1P 3M 5P 7m 9M","dominant ninth","9"],["1P 3M 5P 7m 9M 13M","dominant thirteenth","13"],["1P 3M 5P 7m 11A","lydian dominant seventh","7#11 7#4"],["1P 3M 5P 7m 9m","dominant flat ninth","7b9"],["1P 3M 5P 7m 9A","dominant sharp ninth","7#9"],["1P 3M 7m 9m","altered","alt7"],["1P 4P 5P","suspended fourth","sus4"],["1P 2M 5P","suspended second","sus2"],["1P 4P 5P 7m","suspended fourth seventh","7sus4"],["1P 5P 7m 9M 11P","eleventh","11"],["1P 4P 5P 7m 9m","suspended fourth flat ninth","b9sus phryg"],["1P 5P","fifth","5"],["1P 3M 5A","augmented","aug + +5"],["1P 3M 5A 7M","augmented seventh","maj7#5 maj7+5 +maj7"],["1P 3M 5P 7M 9M 11A","major sharp eleventh (lydian)","maj9#11 Δ9#11"],["1P 2M 4P 5P","","sus24 sus4add9"],["1P 3M 13m","","Mb6"],["1P 3M 5A 7M 9M","","maj9#5 Maj9#5"],["1P 3M 5A 7m","","7#5 +7 7aug aug7"],["1P 3M 5A 7m 9A","","7#5#9 7alt"],["1P 3M 5A 7m 9M","","9#5 9+"],["1P 3M 5A 7m 9M 11A","","9#5#11"],["1P 3M 5A 7m 9m","","7#5b9"],["1P 3M 5A 7m 9m 11A","","7#5b9#11"],["1P 3M 5A 9A","","+add#9"],["1P 3M 5A 9M","","M#5add9 +add9"],["1P 3M 5P 6M 11A","","M6#11 M6b5 6#11 6b5"],["1P 3M 5P 6M 7M 9M","","M7add13"],["1P 3M 5P 6M 9M 11A","","69#11"],["1P 3M 5P 6m 7m","","7b6"],["1P 3M 5P 7M 9A 11A","","maj7#9#11"],["1P 3M 5P 7M 9M 11A 13M","","M13#11 maj13#11 M13+4 M13#4"],["1P 3M 5P 7M 9m","","M7b9"],["1P 3M 5P 7m 11A 13m","","7#11b13 7b5b13"],["1P 3M 5P 7m 13M","","7add6 67 7add13"],["1P 3M 5P 7m 9A 11A","","7#9#11 7b5#9"],["1P 3M 5P 7m 9A 11A 13M","","13#9#11"],["1P 3M 5P 7m 9A 11A 13m","","7#9#11b13"],["1P 3M 5P 7m 9A 13M","","13#9"],["1P 3M 5P 7m 9A 13m","","7#9b13"],["1P 3M 5P 7m 9M 11A","","9#11 9+4 9#4"],["1P 3M 5P 7m 9M 11A 13M","","13#11 13+4 13#4"],["1P 3M 5P 7m 9M 11A 13m","","9#11b13 9b5b13"],["1P 3M 5P 7m 9m 11A","","7b9#11 7b5b9"],["1P 3M 5P 7m 9m 11A 13M","","13b9#11"],["1P 3M 5P 7m 9m 11A 13m","","7b9b13#11 7b9#11b13 7b5b9b13"],["1P 3M 5P 7m 9m 13M","","13b9"],["1P 3M 5P 7m 9m 13m","","7b9b13"],["1P 3M 5P 7m 9m 9A","","7b9#9"],["1P 3M 5P 9M","","Madd9 2 add9 add2"],["1P 3M 5P 9m","","Maddb9"],["1P 3M 5d","","Mb5"],["1P 3M 5d 6M 7m 9M","","13b5"],["1P 3M 5d 7M","","M7b5"],["1P 3M 5d 7M 9M","","M9b5"],["1P 3M 5d 7m","","7b5"],["1P 3M 5d 7m 9M","","9b5"],["1P 3M 7m","","7no5"],["1P 3M 7m 13m","","7b13"],["1P 3M 7m 9M","","9no5"],["1P 3M 7m 9M 13M","","13no5"],["1P 3M 7m 9M 13m","","9b13"],["1P 3m 4P 5P","","madd4"],["1P 3m 5A","","m#5 m+ mb6"],["1P 3m 5P 6M 9M","","m69"],["1P 3m 5P 6m 7M","","mMaj7b6"],["1P 3m 5P 6m 7M 9M","","mMaj9b6"],["1P 3m 5P 7M 9M","","mMaj9"],["1P 3m 5P 7m 11P","","m7add11 m7add4"],["1P 3m 5P 9M","","madd9"],["1P 3m 5d 6M 7M","","o7M7"],["1P 3m 5d 7M","","oM7"],["1P 3m 6m 7M","","mb6M7"],["1P 3m 6m 7m","","m7#5"],["1P 3m 6m 7m 9M","","m9#5"],["1P 3m 6m 7m 9M 11P","","m11A"],["1P 3m 6m 9m","","mb6b9"],["1P 2M 3m 5d 7m","","m9b5"],["1P 4P 5A 7M","","M7#5sus4"],["1P 4P 5A 7M 9M","","M9#5sus4"],["1P 4P 5A 7m","","7#5sus4"],["1P 4P 5P 7M","","M7sus4"],["1P 4P 5P 7M 9M","","M9sus4"],["1P 4P 5P 7m 9M","","9sus4 9sus"],["1P 4P 5P 7m 9M 13M","","13sus4 13sus"],["1P 4P 5P 7m 9m 13m","","7sus4b9b13 7b9b13sus4"],["1P 4P 7m 10m","","4 quartal"],["1P 5P 7m 9m 11P","","11b9"]].forEach(([t,n,e])=>ct(t.split(" "),e.split(" "),n)),ot.sort((t,n)=>t.setNum-n.setNum);var ut={names:function(){return ot.map(t=>t.name).filter(t=>t)},symbols:function(){return ot.map(t=>t.aliases[0]).filter(t=>t)},get:at,all:it,add:ct,removeAll:function(){ot=[],rt={}},keys:function(){return Object.keys(rt)},entries:st,chordType:mt};const lt={weight:0,name:""};const Pt={...R,intervals:[],aliases:[]};let dt=[],Mt={};function pt(){return dt.map(t=>t.name)}function ft(t){return Mt[t]||Pt}const ht=n("ScaleDictionary.scaleType","ScaleType.get",ft);function yt(){return dt.slice()}const bt=n("ScaleDictionary.entries","ScaleType.all",yt);function gt(t,n,e=[]){const o={...H(t),name:n,intervals:t,aliases:e};return dt.push(o),Mt[o.name]=o,Mt[o.setNum]=o,Mt[o.chroma]=o,o.aliases.forEach(t=>function(t,n){Mt[n]=t}(o,t)),o}[["1P 2M 3M 5P 6M","major pentatonic","pentatonic"],["1P 3M 4P 5P 7M","ionian pentatonic"],["1P 3M 4P 5P 7m","mixolydian pentatonic","indian"],["1P 2M 4P 5P 6M","ritusen"],["1P 2M 4P 5P 7m","egyptian"],["1P 3M 4P 5d 7m","neopolitan major pentatonic"],["1P 3m 4P 5P 6m","vietnamese 1"],["1P 2m 3m 5P 6m","pelog"],["1P 2m 4P 5P 6m","kumoijoshi"],["1P 2M 3m 5P 6m","hirajoshi"],["1P 2m 4P 5d 7m","iwato"],["1P 2m 4P 5P 7m","in-sen"],["1P 3M 4A 5P 7M","lydian pentatonic","chinese"],["1P 3m 4P 6m 7m","malkos raga"],["1P 3m 4P 5d 7m","locrian pentatonic","minor seven flat five pentatonic"],["1P 3m 4P 5P 7m","minor pentatonic","vietnamese 2"],["1P 3m 4P 5P 6M","minor six pentatonic"],["1P 2M 3m 5P 6M","flat three pentatonic","kumoi"],["1P 2M 3M 5P 6m","flat six pentatonic"],["1P 2m 3M 5P 6M","scriabin"],["1P 3M 5d 6m 7m","whole tone pentatonic"],["1P 3M 4A 5A 7M","lydian #5P pentatonic"],["1P 3M 4A 5P 7m","lydian dominant pentatonic"],["1P 3m 4P 5P 7M","minor #7M pentatonic"],["1P 3m 4d 5d 7m","super locrian pentatonic"],["1P 2M 3m 4P 5P 7M","minor hexatonic"],["1P 2A 3M 5P 5A 7M","augmented"],["1P 2M 3m 3M 5P 6M","major blues"],["1P 2M 4P 5P 6M 7m","piongio"],["1P 2m 3M 4A 6M 7m","prometheus neopolitan"],["1P 2M 3M 4A 6M 7m","prometheus"],["1P 2m 3M 5d 6m 7m","mystery #1"],["1P 2m 3M 4P 5A 6M","six tone symmetric"],["1P 2M 3M 4A 5A 7m","whole tone","messiaen's mode #1"],["1P 2m 4P 4A 5P 7M","messiaen's mode #5"],["1P 3m 4P 5d 5P 7m","minor blues","blues"],["1P 2M 3M 4P 5d 6m 7m","locrian major","arabian"],["1P 2m 3M 4A 5P 6m 7M","double harmonic lydian"],["1P 2M 3m 4P 5P 6m 7M","harmonic minor"],["1P 2m 3m 3M 5d 6m 7m","altered","super locrian","diminished whole tone","pomeroy"],["1P 2M 3m 4P 5d 6m 7m","locrian #2","half-diminished","aeolian b5"],["1P 2M 3M 4P 5P 6m 7m","mixolydian b6","melodic minor fifth mode","hindu"],["1P 2M 3M 4A 5P 6M 7m","lydian dominant","lydian b7","overtone"],["1P 2M 3M 4A 5P 6M 7M","lydian"],["1P 2M 3M 4A 5A 6M 7M","lydian augmented"],["1P 2m 3m 4P 5P 6M 7m","dorian b2","phrygian #6","melodic minor second mode"],["1P 2M 3m 4P 5P 6M 7M","melodic minor"],["1P 2m 3m 4P 5d 6m 7m","locrian"],["1P 2m 3m 4d 5d 6m 7d","ultralocrian","superlocrian bb7","·superlocrian diminished"],["1P 2m 3m 4P 5d 6M 7m","locrian 6","locrian natural 6","locrian sharp 6"],["1P 2A 3M 4P 5P 5A 7M","augmented heptatonic"],["1P 2M 3m 5d 5P 6M 7m","romanian minor"],["1P 2M 3m 4A 5P 6M 7m","dorian #4"],["1P 2M 3m 4A 5P 6M 7M","lydian diminished"],["1P 2m 3m 4P 5P 6m 7m","phrygian"],["1P 2M 3M 4A 5A 7m 7M","leading whole tone"],["1P 2M 3M 4A 5P 6m 7m","lydian minor"],["1P 2m 3M 4P 5P 6m 7m","phrygian dominant","spanish","phrygian major"],["1P 2m 3m 4P 5P 6m 7M","balinese"],["1P 2m 3m 4P 5P 6M 7M","neopolitan major"],["1P 2M 3m 4P 5P 6m 7m","aeolian","minor"],["1P 2M 3M 4P 5P 6m 7M","harmonic major"],["1P 2m 3M 4P 5P 6m 7M","double harmonic major","gypsy"],["1P 2M 3m 4P 5P 6M 7m","dorian"],["1P 2M 3m 4A 5P 6m 7M","hungarian minor"],["1P 2A 3M 4A 5P 6M 7m","hungarian major"],["1P 2m 3M 4P 5d 6M 7m","oriental"],["1P 2m 3m 3M 4A 5P 7m","flamenco"],["1P 2m 3m 4A 5P 6m 7M","todi raga"],["1P 2M 3M 4P 5P 6M 7m","mixolydian","dominant"],["1P 2m 3M 4P 5d 6m 7M","persian"],["1P 2M 3M 4P 5P 6M 7M","major","ionian"],["1P 2m 3M 5d 6m 7m 7M","enigmatic"],["1P 2M 3M 4P 5A 6M 7M","major augmented","major #5","ionian augmented","ionian #5"],["1P 2A 3M 4A 5P 6M 7M","lydian #9"],["1P 2m 2M 4P 4A 5P 6m 7M","messiaen's mode #4"],["1P 2m 3M 4P 4A 5P 6m 7M","purvi raga"],["1P 2m 3m 3M 4P 5P 6m 7m","spanish heptatonic"],["1P 2M 3M 4P 5P 6M 7m 7M","bebop"],["1P 2M 3m 3M 4P 5P 6M 7m","bebop minor"],["1P 2M 3M 4P 5P 5A 6M 7M","bebop major"],["1P 2m 3m 4P 5d 5P 6m 7m","bebop locrian"],["1P 2M 3m 4P 5P 6m 7m 7M","minor bebop"],["1P 2M 3m 4P 5d 6m 6M 7M","diminished","whole-half diminished"],["1P 2M 3M 4P 5d 5P 6M 7M","ichikosucho"],["1P 2M 3m 4P 5P 6m 6M 7M","minor six diminished"],["1P 2m 3m 3M 4A 5P 6M 7m","half-whole diminished","dominant diminished","messiaen's mode #2"],["1P 3m 3M 4P 5P 6M 7m 7M","kafi raga"],["1P 2M 3M 4P 4A 5A 6A 7M","messiaen's mode #6"],["1P 2M 3m 3M 4P 5d 5P 6M 7m","composite blues"],["1P 2M 3m 3M 4A 5P 6m 7m 7M","messiaen's mode #3"],["1P 2m 2M 3m 4P 4A 5P 6m 6M 7M","messiaen's mode #7"],["1P 2m 2M 3m 3M 4P 5d 5P 6m 6M 7m 7M","chromatic"]].forEach(([t,n,...e])=>gt(t.split(" "),n,e));var At={names:pt,get:ft,all:yt,add:gt,removeAll:function(){dt=[],Mt={}},keys:function(){return Object.keys(Mt)},entries:bt,scaleType:ht};const vt={empty:!0,name:"",symbol:"",root:"",rootDegree:0,type:"",tonic:null,setNum:NaN,quality:"Unknown",chroma:"",normalized:"",aliases:[],notes:[],intervals:[]},jt=/^(6|64|7|9|11|13)$/;function It(t){const[n,e,o,r]=f(t);return""===n?["",t]:"A"===n&&"ug"===r?["","aug"]:r||"4"!==o&&"5"!==o?jt.test(o)?[n+e,o+r]:[n+e+o,r]:[n+e,o]}function wt(t){if(""===t)return vt;if(Array.isArray(t)&&2===t.length)return Ot(t[1],t[0]);{const[n,e]=It(t),o=Ot(e,n);return o.empty?Ot(t):o}}function Ot(t,n,e){const o=at(t),r=M(n||""),a=M(e||"");if(o.empty||n&&r.empty||e&&a.empty)return vt;const m=N(r.pc,a.pc),i=o.intervals.indexOf(m)+1;if(!a.empty&&!i)return vt;const s=r.empty?[]:o.intervals.map(t=>O(r,t));t=-1!==o.aliases.indexOf(t)?t:o.aliases[0];const c=`${r.empty?"":r.pc}${t}${a.empty?"":"/"+a.pc}`,u=`${n?r.pc+" ":""}${o.name}${e?" over "+a.pc:""}`;return{...o,name:u,symbol:c,type:o.name,root:a.name,rootDegree:i,tonic:r.name,notes:s}}var Nt={getChord:Ot,get:wt,detect:function(t){const n=t.map(t=>M(t).pc).filter(t=>t);return 0===M.length?[]:function(t,n){const e=t[0],o=M(e).chroma,r=(t=>{const n=t.reduce((t,n)=>{const e=M(n).chroma;return void 0!==e&&(t[e]=t[e]||M(n).name),t},{});return t=>n[t]})(t);return X(t,!1).map((t,a)=>{const m=at(t).aliases[0];if(!m)return lt;const i=r(a);return a!==o?{weight:.5*n,name:`${i}${m}/${e}`}:{weight:1*n,name:`${i}${m}`}})}(n,1).filter(t=>t.weight).sort((t,n)=>n.weight-t.weight).map(t=>t.name)},chordScales:function(t){const n=Z(wt(t).chroma);return yt().filter(t=>n(t.chroma)).map(t=>t.name)},extended:function(t){const n=wt(t),e=Z(n.chroma);return it().filter(t=>e(t.chroma)).map(t=>n.tonic+t.aliases[0])},reduced:function(t){const n=wt(t),e=Y(n.chroma);return it().filter(t=>e(t.chroma)).map(t=>n.tonic+t.aliases[0])},tokenize:It,transpose:function(t,n){const[e,o]=It(t);return e?O(e,n)+o:t},chord:n("Chord.chord","Chord.get",wt)};const xt=[];[[.125,"dl",["large","duplex longa","maxima","octuple","octuple whole"]],[.25,"l",["long","longa"]],[.5,"d",["double whole","double","breve"]],[1,"w",["whole","semibreve"]],[2,"h",["half","minim"]],[4,"q",["quarter","crotchet"]],[8,"e",["eighth","quaver"]],[16,"s",["sixteenth","semiquaver"]],[32,"t",["thirty-second","demisemiquaver"]],[64,"sf",["sixty-fourth","hemidemisemiquaver"]],[128,"h",["hundred twenty-eighth"]],[256,"th",["two hundred fifty-sixth"]]].forEach(([t,n,e])=>function(t,n,e){xt.push({empty:!1,dots:"",name:"",value:1/t,fraction:t<1?[1/t,1]:[1,t],shorthand:n,names:e})}(t,n,e));const Tt={empty:!0,name:"",value:0,fraction:[0,0],shorthand:"",dots:"",names:[]};const Dt=/^([^.]+)(\.*)$/;function St(t){const[n,e,o]=Dt.exec(t)||[],r=xt.find(t=>t.shorthand===e||t.names.includes(e));if(!r)return Tt;const a=function(t,n){const e=Math.pow(2,n);let o=t[0]*e,r=t[1]*e;const a=o;for(let t=0;t<n;t++)o+=a/Math.pow(2,t+1);for(;o%2==0&&r%2==0;)o/=2,r/=2;return[o,r]}(r.fraction,o.length),m=a[0]/a[1];return{...r,name:t,dots:o,value:m,fraction:a}}var Vt={names:function(){return xt.reduce((t,n)=>(n.names.forEach(n=>t.push(n)),t),[])},shorthands:function(){return xt.map(t=>t.shorthand)},get:St,value:t=>St(t).value,fraction:t=>St(t).fraction};const Ct=j;const $t=[1,2,2,3,3,4,5,5,6,6,7,7],qt="P m M m M P d P m M m M".split(" ");const kt=N,Et=zt((t,n)=>[t[0]+n[0],t[1]+n[1]]),_t=zt((t,n)=>[t[0]-n[0],t[1]-n[1]]);var Ft={names:function(){return"1P 2M 3M 4P 5P 6m 7m".split(" ")},get:Ct,name:t=>j(t).name,num:t=>j(t).num,semitones:t=>j(t).semitones,quality:t=>j(t).q,fromSemitones:function(t){const n=t<0?-1:1,e=Math.abs(t),o=e%12,r=Math.floor(e/12);return n*($t[o]+7*r)+qt[o]},distance:kt,invert:function(t){const n=j(t);return n.empty?"":j({step:(7-n.step)%7,alt:"perfectable"===n.type?-n.alt:-(n.alt+1),oct:n.oct,dir:n.dir}).name},simplify:function(t){const n=j(t);return n.empty?"":n.simple+n.q},add:Et,addTo:t=>n=>Et(t,n),substract:_t};function zt(t){return(n,e)=>{const o=j(n).coord,r=j(e).coord;if(o&&r){return w(t(o,r)).name}}}function Rt(t){return+t>=0&&+t<=127}function Ut(t){if(Rt(t))return+t;const n=M(t);return n.empty?null:n.midi}const Bt=Math.log(2),Gt=Math.log(440);const Kt="C C# D D# E F F# G G# A A# B".split(" "),Lt="C Db D Eb E F Gb G Ab A Bb B".split(" ");function Ht(t,n={}){t=Math.round(t);const e=(!0===n.sharps?Kt:Lt)[t%12];return n.pitchClass?e:e+(Math.floor(t/12)-1)}var Jt={isMidi:Rt,toMidi:Ut,midiToFreq:function(t,n=440){return Math.pow(2,(t-69)/12)*n},midiToNoteName:Ht,freqToMidi:function(t){const n=12*(Math.log(t)-Gt)/Bt+69;return Math.round(100*n)/100}};const Qt=["C","D","E","F","G","A","B"],Wt=t=>t.name,Xt=t=>t.map(M).filter(t=>!t.empty);const Yt=M;const Zt=O,tn=O,nn=t=>n=>Zt(n,t),en=nn,on=t=>n=>Zt(t,n),rn=on;function an(t,n){const e=Yt(t);if(e.empty)return"";const[o,r]=e.coord;return h(void 0===r?[o+n]:[o+n,r]).name}const mn=an,sn=(t,n)=>t.height-n.height;function cn(t,n){return n=n||sn,Xt(t).sort(n).map(Wt)}function un(t){return cn(t,sn).filter((t,n,e)=>0===n||t!==e[n-1])}const ln=dn(!0),Pn=dn(!1);function dn(t){return n=>{const e=Yt(n);if(e.empty)return"";const o=t?e.alt>0:e.alt<0,r=null===e.midi;return Ht(e.midi||e.chroma,{sharps:o,pitchClass:r})}}var Mn={names:function(t){return void 0===t?Qt.slice():Array.isArray(t)?Xt(t).map(Wt):[]},get:Yt,name:t=>Yt(t).name,pitchClass:t=>Yt(t).pc,accidentals:t=>Yt(t).acc,octave:t=>Yt(t).oct,midi:t=>Yt(t).midi,ascending:sn,descending:(t,n)=>n.height-t.height,sortedNames:cn,sortedUniqNames:un,fromMidi:function(t){return Ht(t)},fromMidiSharps:function(t){return Ht(t,{sharps:!0})},freq:t=>Yt(t).freq,chroma:t=>Yt(t).chroma,transpose:Zt,tr:tn,transposeBy:nn,trBy:en,transposeFrom:on,trFrom:rn,transposeFifths:an,trFifths:mn,simplify:ln,enharmonic:Pn};const pn={empty:!0,name:"",chordType:""},fn={};function hn(t){return"string"==typeof t?fn[t]||(fn[t]=function(t){const[n,e,o,r]=(a=t,bn.exec(a)||["","","",""]);var a;if(!o)return pn;const m=o.toUpperCase(),i=An.indexOf(m),s=d(e);return{empty:!1,name:n,roman:o,interval:j({step:i,alt:s,dir:1}).name,acc:e,chordType:r,alt:s,step:i,major:o===m,oct:0,dir:1}}(t)):"number"==typeof t?hn(An[t]||""):o(t)?hn(P((n=t).alt)+An[n.step]):e(t)?hn(t.name):pn;var n}const yn=n("RomanNumeral.romanNumeral","RomanNumeral.get",hn);const bn=/^(#{1,}|b{1,}|x{1,}|)(IV|I{1,3}|VI{0,2}|iv|i{1,3}|vi{0,2})([^IViv]*)$/;const gn="I II III IV V VI VII",An=gn.split(" "),vn=gn.toLowerCase().split(" ");var jn={names:function(t=!0){return(t?An:vn).slice()},get:hn,romanNumeral:yn};const In=Object.freeze([]),wn={type:"major",tonic:"",alteration:0,keySignature:""},On={tonic:"",grades:In,intervals:In,scale:In,chords:In,chordsHarmonicFunction:In,chordScales:In},Nn={...wn,...On,type:"major",minorRelative:"",scale:In,secondaryDominants:In,secondaryDominantsMinorRelative:In,substituteDominants:In,substituteDominantsMinorRelative:In},xn={...wn,type:"minor",relativeMajor:"",natural:On,harmonic:On,melodic:On},Tn=(t,n,e="")=>n.map((n,o)=>`${t[o]}${e}${n}`);function Dn(t,n,e,o){return r=>{const a=t.map(t=>hn(t).interval||""),m=a.map(t=>O(r,t));return{tonic:r,grades:t,intervals:a,scale:m,chords:Tn(m,n),chordsHarmonicFunction:Tn(m,e),chordScales:Tn(m,o," ")}}}const Sn=(t,n)=>{const e=M(t),o=M(n);return e.empty||o.empty?0:o.coord[0]-e.coord[0]},Vn=Dn("I II III IV V VI VII".split(" "),"maj7 m7 m7 maj7 7 m7 m7b5".split(" "),"T SD T SD D T D".split(" "),"major,dorian,phrygian,lydian,mixolydian,minor,locrian".split(",")),Cn=Dn("I II bIII IV V bVI bVII".split(" "),"m7 m7b5 maj7 m7 m7 maj7 7".split(" "),"T SD T SD D SD SD".split(" "),"minor,locrian,major,dorian,phrygian,lydian,mixolydian".split(",")),$n=Dn("I II bIII IV V bVI VII".split(" "),"mMaj7 m7b5 +maj7 m7 7 maj7 o7".split(" "),"T SD T SD D SD D".split(" "),"harmonic minor,locrian 6,major augmented,lydian diminished,phrygian dominant,lydian #9,ultralocrian".split(",")),qn=Dn("I II bIII IV V VI VII".split(" "),"m6 m7 +maj7 7 7 m7b5 m7b5".split(" "),"T SD T SD D - -".split(" "),"melodic minor,dorian b2,lydian augmented,lydian dominant,mixolydian b6,locrian #2,altered".split(","));var kn={majorKey:function(t){const n=M(t).pc;if(!n)return Nn;const e=Vn(n),o=Sn("C",n),r=n=>{const e=hn(n);return e.empty?"":O(t,e.interval)+e.chordType};return{...e,type:"major",minorRelative:O(n,"-3m"),alteration:o,keySignature:P(o),secondaryDominants:"- VI7 VII7 I7 II7 III7 -".split(" ").map(r),secondaryDominantsMinorRelative:"- IIIm7b5 IV#m7 Vm7 VIm7 VIIm7b5 -".split(" ").map(r),substituteDominants:"- bIII7 IV7 bV7 bVI7 bVII7 -".split(" ").map(r),substituteDominantsMinorRelative:"- IIIm7 Im7 IIbm7 VIm7 IVm7 -".split(" ").map(r)}},majorTonicFromKeySignature:function(t){return"number"==typeof t?an("C",t):"string"==typeof t&&/^b+|#+$/.test(t)?an("C",d(t)):null},minorKey:function(t){const n=M(t).pc;if(!n)return xn;const e=Sn("C",n)-3;return{type:"minor",tonic:n,relativeMajor:O(n,"3m"),alteration:e,keySignature:P(e),natural:Cn(n),harmonic:$n(n),melodic:qn(n)}}};const En={...R,name:"",alt:0,modeNum:NaN,triad:"",seventh:"",aliases:[]},_n=[[0,2773,0,"ionian","","Maj7","major"],[1,2902,2,"dorian","m","m7"],[2,3418,4,"phrygian","m","m7"],[3,2741,-1,"lydian","","Maj7"],[4,2774,1,"mixolydian","","7"],[5,2906,3,"aeolian","m","m7","minor"],[6,3434,5,"locrian","dim","m7b5"]].map((function(t){const[n,e,o,r,a,m,i]=t,s=i?[i]:[],c=Number(e).toString(2);return{empty:!1,intervals:W(c),modeNum:n,chroma:c,normalized:c,name:r,setNum:e,alt:o,triad:a,seventh:m,aliases:s}})),Fn={};function zn(t){return"string"==typeof t?Fn[t.toLowerCase()]||En:t&&t.name?zn(t.name):En}_n.forEach(t=>{Fn[t.name]=t,t.aliases.forEach(n=>{Fn[n]=t})});const Rn=n("Mode.mode","Mode.get",zn);function Un(){return _n.slice()}var Bn={get:zn,names:function(){return _n.map(t=>t.name)},all:Un,entries:n("Mode.mode","Mode.all",Un),mode:Rn};var Gn={fromRomanNumerals:function(t,n){return n.map(hn).map(n=>O(t,j(n))+n.chordType)},toRomanNumerals:function(t,n){return n.map(n=>{const[e,o]=It(n);return hn(j(N(t,e))).name+o})}};function Kn(t){const n=F(t.map(Ut));return t.length&&n.length===t.length?n.reduce((t,n)=>{const e=t[t.length-1];return t.concat(E(e,n).slice(1))},[n[0]]):[]}var Ln={numeric:Kn,chromatic:function(t,n){return Kn(t).map(t=>Ht(t,n))}};const Hn={empty:!0,name:"",type:"",tonic:null,setNum:NaN,chroma:"",normalized:"",aliases:[],notes:[],intervals:[]};function Jn(t){if("string"!=typeof t)return["",""];const n=t.indexOf(" "),e=M(t.substring(0,n));if(e.empty){const n=M(t);return n.empty?["",t]:[n.name,""]}const o=t.substring(e.name.length+1);return[e.name,o.length?o:""]}function Qn(t){const n=Array.isArray(t)?t:Jn(t),e=M(n[0]).name,o=ft(n[1]);if(o.empty)return Hn;const r=o.name,a=e?o.intervals.map(t=>O(e,t)):[],m=e?e+" "+r:r;return{...o,name:m,type:r,tonic:e,notes:a}}var Wn={get:Qn,names:pt,extended:function(t){const n=Z(Qn(t).chroma);return yt().filter(t=>n(t.chroma)).map(t=>t.name)},modeNames:function(t){const n=Qn(t);if(n.empty)return[];const e=n.tonic?n.notes:n.intervals;return X(n.chroma).map((t,n)=>{const o=Qn(t).name;return o?[e[n],o]:["",""]}).filter(t=>t[0])},reduced:function(t){const n=Y(Qn(t).chroma);return yt().filter(t=>n(t.chroma)).map(t=>t.name)},scaleChords:function(t){const n=Y(Qn(t).chroma);return it().filter(t=>n(t.chroma)).map(t=>t.aliases[0])},scaleNotes:function(t){const n=t.map(t=>M(t).pc).filter(t=>t),e=n[0],o=un(n);return _(o.indexOf(e),o)},tokenize:Jn,scale:n("Scale.scale","Scale.get",Qn)};const Xn={empty:!0,name:"",upper:void 0,lower:void 0,type:void 0,additive:[]},Yn=["4/4","3/4","2/4","2/2","12/8","9/8","6/8","3/8"];const Zn=/^(\d?\d(?:\+\d)*)\/(\d)$/,te=new Map;function ne(t){if("string"==typeof t){const[n,e,o]=Zn.exec(t)||[];return ne([e,o])}const[n,e]=t,o=+e;if("number"==typeof n)return[n,o];const r=n.split("+").map(t=>+t);return 1===r.length?[r[0],o]:[r,o]}var ee={names:function(){return Yn.slice()},parse:ne,get:function(t){const n=te.get(t);if(n)return n;const e=function([t,n]){const e=Array.isArray(t)?t.reduce((t,n)=>t+n,0):t,o=n;if(0===e||0===o)return Xn;const r=Array.isArray(t)?`${t.join("+")}/${n}`:`${t}/${n}`,a=Array.isArray(t)?t:[];return{empty:!1,name:r,type:4===o||2===o?"simple":8===o&&e%3==0?"compound":"irregular",upper:e,lower:o,additive:a}}(ne(t));return te.set(t,e),e}};var oe,re,ae=(function(t,n){!function(t,n,e,o,r,a,m,i,s,c,u,l,P,d,M,p,f,h,y,b){n=n&&Object.prototype.hasOwnProperty.call(n,"default")?n.default:n,o=o&&Object.prototype.hasOwnProperty.call(o,"default")?o.default:o,r=r&&Object.prototype.hasOwnProperty.call(r,"default")?r.default:r,a=a&&Object.prototype.hasOwnProperty.call(a,"default")?a.default:a,i=i&&Object.prototype.hasOwnProperty.call(i,"default")?i.default:i,s=s&&Object.prototype.hasOwnProperty.call(s,"default")?s.default:s,c=c&&Object.prototype.hasOwnProperty.call(c,"default")?c.default:c,u=u&&Object.prototype.hasOwnProperty.call(u,"default")?u.default:u,l=l&&Object.prototype.hasOwnProperty.call(l,"default")?l.default:l,P=P&&Object.prototype.hasOwnProperty.call(P,"default")?P.default:P,d=d&&Object.prototype.hasOwnProperty.call(d,"default")?d.default:d,M=M&&Object.prototype.hasOwnProperty.call(M,"default")?M.default:M,p=p&&Object.prototype.hasOwnProperty.call(p,"default")?p.default:p,f=f&&Object.prototype.hasOwnProperty.call(f,"default")?f.default:f,h=h&&Object.prototype.hasOwnProperty.call(h,"default")?h.default:h,y=y&&Object.prototype.hasOwnProperty.call(y,"default")?y.default:y,b=b&&Object.prototype.hasOwnProperty.call(b,"default")?b.default:b;var g=m,A=d,v=r,j=y;Object.keys(m).forEach((function(n){"default"!==n&&Object.defineProperty(t,n,{enumerable:!0,get:function(){return m[n]}})})),t.AbcNotation=n,t.Array=e,t.Chord=o,t.ChordType=r,t.Collection=a,t.Core=m,t.DurationValue=i,t.Interval=s,t.Key=c,t.Midi=u,t.Mode=l,t.Note=P,t.Pcset=d,t.Progression=M,t.Range=p,t.RomanNumeral=f,t.Scale=h,t.ScaleType=y,t.TimeSignature=b,t.ChordDictionary=v,t.PcSet=A,t.ScaleDictionary=j,t.Tonal=g,Object.defineProperty(t,"__esModule",{value:!0})}(n,$,k,Nt,ut,z,x,Vt,Ft,kn,Jt,Bn,Mn,nt,Gn,Ln,jn,Wn,At,ee)}(oe={exports:{}},oe.exports),oe.exports);return(re=ae)&&re.__esModule&&Object.prototype.hasOwnProperty.call(re,"default")?re.default:re}();
)END";





} // JsLibraries
