
#include "Matrix88M.hpp"
#include "mixer.hpp"

void Matrix88M::processAlways(const ProcessArgs& args) {
	int position = 0;
	int baseID = 0;
	if (baseConnected()) {
		Matrix88ExpanderMessage* bm = fromBase();
		baseID = bm->baseID;
		position = bm->position;
	}
	setBaseIDAndPosition(baseID, position);

	if (expanderConnected()) {
		Matrix88ExpanderMessage* te = toExpander();
		te->baseID = _baseID;
		te->position = _position > 0 ? _position + 1 : 0;
	}
}

struct Matrix88MWidget : MutesMatrixExpanderModuleWidget {
	static constexpr int hp = 20;

	Matrix88MWidget(Matrix88M* module) {
		setModule(module);
		box.size = Vec(RACK_GRID_WIDTH * hp, RACK_GRID_HEIGHT);
		setPanel(box.size, "Matrix88M");
		createScrews();

		// generated by svg_widgets.rb
		auto mute11ParamPosition = Vec(18.5, 33.0);
		auto mute21ParamPosition = Vec(18.5, 70.0);
		auto mute31ParamPosition = Vec(18.5, 107.0);
		auto mute41ParamPosition = Vec(18.5, 144.0);
		auto mute51ParamPosition = Vec(18.5, 181.0);
		auto mute61ParamPosition = Vec(18.5, 218.0);
		auto mute71ParamPosition = Vec(18.5, 255.0);
		auto mute81ParamPosition = Vec(18.5, 292.0);
		auto mute12ParamPosition = Vec(53.5, 33.0);
		auto mute22ParamPosition = Vec(53.5, 70.0);
		auto mute32ParamPosition = Vec(53.5, 107.0);
		auto mute42ParamPosition = Vec(53.5, 144.0);
		auto mute52ParamPosition = Vec(53.5, 181.0);
		auto mute62ParamPosition = Vec(53.5, 218.0);
		auto mute72ParamPosition = Vec(53.5, 255.0);
		auto mute82ParamPosition = Vec(53.5, 292.0);
		auto mute13ParamPosition = Vec(88.5, 33.0);
		auto mute23ParamPosition = Vec(88.5, 70.0);
		auto mute33ParamPosition = Vec(88.5, 107.0);
		auto mute43ParamPosition = Vec(88.5, 144.0);
		auto mute53ParamPosition = Vec(88.5, 181.0);
		auto mute63ParamPosition = Vec(88.5, 218.0);
		auto mute73ParamPosition = Vec(88.5, 255.0);
		auto mute83ParamPosition = Vec(88.5, 292.0);
		auto mute14ParamPosition = Vec(123.5, 33.0);
		auto mute24ParamPosition = Vec(123.5, 70.0);
		auto mute34ParamPosition = Vec(123.5, 107.0);
		auto mute44ParamPosition = Vec(123.5, 144.0);
		auto mute54ParamPosition = Vec(123.5, 181.0);
		auto mute64ParamPosition = Vec(123.5, 218.0);
		auto mute74ParamPosition = Vec(123.5, 255.0);
		auto mute84ParamPosition = Vec(123.5, 292.0);
		auto mute15ParamPosition = Vec(158.5, 33.0);
		auto mute25ParamPosition = Vec(158.5, 70.0);
		auto mute35ParamPosition = Vec(158.5, 107.0);
		auto mute45ParamPosition = Vec(158.5, 144.0);
		auto mute55ParamPosition = Vec(158.5, 181.0);
		auto mute65ParamPosition = Vec(158.5, 218.0);
		auto mute75ParamPosition = Vec(158.5, 255.0);
		auto mute85ParamPosition = Vec(158.5, 292.0);
		auto mute16ParamPosition = Vec(193.5, 33.0);
		auto mute26ParamPosition = Vec(193.5, 70.0);
		auto mute36ParamPosition = Vec(193.5, 107.0);
		auto mute46ParamPosition = Vec(193.5, 144.0);
		auto mute56ParamPosition = Vec(193.5, 181.0);
		auto mute66ParamPosition = Vec(193.5, 218.0);
		auto mute76ParamPosition = Vec(193.5, 255.0);
		auto mute86ParamPosition = Vec(193.5, 292.0);
		auto mute17ParamPosition = Vec(228.5, 33.0);
		auto mute27ParamPosition = Vec(228.5, 70.0);
		auto mute37ParamPosition = Vec(228.5, 107.0);
		auto mute47ParamPosition = Vec(228.5, 144.0);
		auto mute57ParamPosition = Vec(228.5, 181.0);
		auto mute67ParamPosition = Vec(228.5, 218.0);
		auto mute77ParamPosition = Vec(228.5, 255.0);
		auto mute87ParamPosition = Vec(228.5, 292.0);
		auto mute18ParamPosition = Vec(263.5, 33.0);
		auto mute28ParamPosition = Vec(263.5, 70.0);
		auto mute38ParamPosition = Vec(263.5, 107.0);
		auto mute48ParamPosition = Vec(263.5, 144.0);
		auto mute58ParamPosition = Vec(263.5, 181.0);
		auto mute68ParamPosition = Vec(263.5, 218.0);
		auto mute78ParamPosition = Vec(263.5, 255.0);
		auto mute88ParamPosition = Vec(263.5, 292.0);
		// end generated by svg_widgets.rb

		addParam(createParam<SoloMuteButton>(mute11ParamPosition, module, Matrix88M::MUTE11_PARAM));
		addParam(createParam<SoloMuteButton>(mute21ParamPosition, module, Matrix88M::MUTE21_PARAM));
		addParam(createParam<SoloMuteButton>(mute31ParamPosition, module, Matrix88M::MUTE31_PARAM));
		addParam(createParam<SoloMuteButton>(mute41ParamPosition, module, Matrix88M::MUTE41_PARAM));
		addParam(createParam<SoloMuteButton>(mute51ParamPosition, module, Matrix88M::MUTE51_PARAM));
		addParam(createParam<SoloMuteButton>(mute61ParamPosition, module, Matrix88M::MUTE61_PARAM));
		addParam(createParam<SoloMuteButton>(mute71ParamPosition, module, Matrix88M::MUTE71_PARAM));
		addParam(createParam<SoloMuteButton>(mute81ParamPosition, module, Matrix88M::MUTE81_PARAM));
		addParam(createParam<SoloMuteButton>(mute12ParamPosition, module, Matrix88M::MUTE12_PARAM));
		addParam(createParam<SoloMuteButton>(mute22ParamPosition, module, Matrix88M::MUTE22_PARAM));
		addParam(createParam<SoloMuteButton>(mute32ParamPosition, module, Matrix88M::MUTE32_PARAM));
		addParam(createParam<SoloMuteButton>(mute42ParamPosition, module, Matrix88M::MUTE42_PARAM));
		addParam(createParam<SoloMuteButton>(mute52ParamPosition, module, Matrix88M::MUTE52_PARAM));
		addParam(createParam<SoloMuteButton>(mute62ParamPosition, module, Matrix88M::MUTE62_PARAM));
		addParam(createParam<SoloMuteButton>(mute72ParamPosition, module, Matrix88M::MUTE72_PARAM));
		addParam(createParam<SoloMuteButton>(mute82ParamPosition, module, Matrix88M::MUTE82_PARAM));
		addParam(createParam<SoloMuteButton>(mute13ParamPosition, module, Matrix88M::MUTE13_PARAM));
		addParam(createParam<SoloMuteButton>(mute23ParamPosition, module, Matrix88M::MUTE23_PARAM));
		addParam(createParam<SoloMuteButton>(mute33ParamPosition, module, Matrix88M::MUTE33_PARAM));
		addParam(createParam<SoloMuteButton>(mute43ParamPosition, module, Matrix88M::MUTE43_PARAM));
		addParam(createParam<SoloMuteButton>(mute53ParamPosition, module, Matrix88M::MUTE53_PARAM));
		addParam(createParam<SoloMuteButton>(mute63ParamPosition, module, Matrix88M::MUTE63_PARAM));
		addParam(createParam<SoloMuteButton>(mute73ParamPosition, module, Matrix88M::MUTE73_PARAM));
		addParam(createParam<SoloMuteButton>(mute83ParamPosition, module, Matrix88M::MUTE83_PARAM));
		addParam(createParam<SoloMuteButton>(mute14ParamPosition, module, Matrix88M::MUTE14_PARAM));
		addParam(createParam<SoloMuteButton>(mute24ParamPosition, module, Matrix88M::MUTE24_PARAM));
		addParam(createParam<SoloMuteButton>(mute34ParamPosition, module, Matrix88M::MUTE34_PARAM));
		addParam(createParam<SoloMuteButton>(mute44ParamPosition, module, Matrix88M::MUTE44_PARAM));
		addParam(createParam<SoloMuteButton>(mute54ParamPosition, module, Matrix88M::MUTE54_PARAM));
		addParam(createParam<SoloMuteButton>(mute64ParamPosition, module, Matrix88M::MUTE64_PARAM));
		addParam(createParam<SoloMuteButton>(mute74ParamPosition, module, Matrix88M::MUTE74_PARAM));
		addParam(createParam<SoloMuteButton>(mute84ParamPosition, module, Matrix88M::MUTE84_PARAM));
		addParam(createParam<SoloMuteButton>(mute15ParamPosition, module, Matrix88M::MUTE15_PARAM));
		addParam(createParam<SoloMuteButton>(mute25ParamPosition, module, Matrix88M::MUTE25_PARAM));
		addParam(createParam<SoloMuteButton>(mute35ParamPosition, module, Matrix88M::MUTE35_PARAM));
		addParam(createParam<SoloMuteButton>(mute45ParamPosition, module, Matrix88M::MUTE45_PARAM));
		addParam(createParam<SoloMuteButton>(mute55ParamPosition, module, Matrix88M::MUTE55_PARAM));
		addParam(createParam<SoloMuteButton>(mute65ParamPosition, module, Matrix88M::MUTE65_PARAM));
		addParam(createParam<SoloMuteButton>(mute75ParamPosition, module, Matrix88M::MUTE75_PARAM));
		addParam(createParam<SoloMuteButton>(mute85ParamPosition, module, Matrix88M::MUTE85_PARAM));
		addParam(createParam<SoloMuteButton>(mute16ParamPosition, module, Matrix88M::MUTE16_PARAM));
		addParam(createParam<SoloMuteButton>(mute26ParamPosition, module, Matrix88M::MUTE26_PARAM));
		addParam(createParam<SoloMuteButton>(mute36ParamPosition, module, Matrix88M::MUTE36_PARAM));
		addParam(createParam<SoloMuteButton>(mute46ParamPosition, module, Matrix88M::MUTE46_PARAM));
		addParam(createParam<SoloMuteButton>(mute56ParamPosition, module, Matrix88M::MUTE56_PARAM));
		addParam(createParam<SoloMuteButton>(mute66ParamPosition, module, Matrix88M::MUTE66_PARAM));
		addParam(createParam<SoloMuteButton>(mute76ParamPosition, module, Matrix88M::MUTE76_PARAM));
		addParam(createParam<SoloMuteButton>(mute86ParamPosition, module, Matrix88M::MUTE86_PARAM));
		addParam(createParam<SoloMuteButton>(mute17ParamPosition, module, Matrix88M::MUTE17_PARAM));
		addParam(createParam<SoloMuteButton>(mute27ParamPosition, module, Matrix88M::MUTE27_PARAM));
		addParam(createParam<SoloMuteButton>(mute37ParamPosition, module, Matrix88M::MUTE37_PARAM));
		addParam(createParam<SoloMuteButton>(mute47ParamPosition, module, Matrix88M::MUTE47_PARAM));
		addParam(createParam<SoloMuteButton>(mute57ParamPosition, module, Matrix88M::MUTE57_PARAM));
		addParam(createParam<SoloMuteButton>(mute67ParamPosition, module, Matrix88M::MUTE67_PARAM));
		addParam(createParam<SoloMuteButton>(mute77ParamPosition, module, Matrix88M::MUTE77_PARAM));
		addParam(createParam<SoloMuteButton>(mute87ParamPosition, module, Matrix88M::MUTE87_PARAM));
		addParam(createParam<SoloMuteButton>(mute18ParamPosition, module, Matrix88M::MUTE18_PARAM));
		addParam(createParam<SoloMuteButton>(mute28ParamPosition, module, Matrix88M::MUTE28_PARAM));
		addParam(createParam<SoloMuteButton>(mute38ParamPosition, module, Matrix88M::MUTE38_PARAM));
		addParam(createParam<SoloMuteButton>(mute48ParamPosition, module, Matrix88M::MUTE48_PARAM));
		addParam(createParam<SoloMuteButton>(mute58ParamPosition, module, Matrix88M::MUTE58_PARAM));
		addParam(createParam<SoloMuteButton>(mute68ParamPosition, module, Matrix88M::MUTE68_PARAM));
		addParam(createParam<SoloMuteButton>(mute78ParamPosition, module, Matrix88M::MUTE78_PARAM));
		addParam(createParam<SoloMuteButton>(mute88ParamPosition, module, Matrix88M::MUTE88_PARAM));
	}
};

Model* modelMatrix88M = createModel<Matrix88M, Matrix88MWidget>("Bogaudio-Matrix88M", "MX88M", "Mutes expander for MATRIX88", "Mixer", "Expander", "Polyphonic");
